import 'dart:convert';

import 'package:flutter/foundation.dart';
import 'package:rec_api_dart/src/entities/schedule/schedule_day.ent.dart';
import 'package:rec_api_dart/src/entities/schedule/schedule_type.dart';
import 'package:rec_api_dart/src/helpers/date_helper.dart';

enum ScheduleState {
  open,
  openAllDay,
  closed,
  notAvailable;

  String get name {
    switch (this) {
      case ScheduleState.open:
        return 'OPEN';
      case ScheduleState.openAllDay:
        return 'OPEN_ALL_DAY';
      case ScheduleState.closed:
        return 'CLOSED';
      case ScheduleState.notAvailable:
        return 'NOT_AVAILABLE';
    }
  }
}

class Schedule {
  /// Specifies the type of schedule
  ScheduleType? type;

  /// Specifies the schedule for each day of the week
  List<ScheduleDay> days;

  Schedule({
    ScheduleType? type,
    List<ScheduleDay>? days,
  })  : type = type ?? ScheduleType.NOT_AVAILABLE,
        days =
            (days != null && days.isNotEmpty) ? days : List.generate(7, (index) => ScheduleDay());

  bool get isNotAvailable => type == ScheduleType.NOT_AVAILABLE;
  bool get isClosed => type == ScheduleType.CLOSED;
  bool get isOpen24h => type == ScheduleType.FULL;
  bool get isDefined => type == ScheduleType.TIMETABLE;

  /// Returns the state of the schedule for the current day
  /// For example: Closed · Opens at 10am
  ScheduleState getStateForDate(DateTime now) {
    if (type == ScheduleType.CLOSED) {
      return ScheduleState.closed;
    }
    if (type == ScheduleType.FULL) {
      return ScheduleState.openAllDay;
    }
    if (type == ScheduleType.TIMETABLE) {
      return isOpenToday(now) ? ScheduleState.open : ScheduleState.closed;
    }

    return ScheduleState.notAvailable;
  }

  String getStateNameForDate(DateTime now) {
    return getStateForDate(now).name;
  }

  /// Returns the schedule for specified [day]
  ScheduleDay getScheduleForDay(int day) {
    return days[day];
  }

  /// Returns the schedule for current day
  ScheduleDay getScheduleForWeekday(DateTime now) {
    return getScheduleForDay(now.weekday - 1);
  }

  /// Returns whether it's opened or closed for current day
  bool isOpenToday(DateTime now) {
    return getScheduleForWeekday(now).isOpen(now);
  }

  DateTime? getNextOpeningTime([DateTime? clock]) {
    final today = (clock ?? DateTime.now());
    final todaySchedule = getScheduleForWeekday(today);
    final closedForTheDay = todaySchedule.isClosedForTheDay(today);

    debugPrint('todays schedule closedForTheDay: $closedForTheDay');

    // It's closed so return next opening date
    if (todaySchedule.isClosedForTheDay(today)) {
      for (var day = 1; day <= 7; day++) {
        debugPrint('day: $day');
        if (day == today.weekday - 1) continue;

        final nextDay = DateTime(
          today.year,
          today.month,
          today.day,
          today.hour,
          today.minute,
        ).add(Duration(days: day));

        final state = getStateForDate(nextDay);
        final weekdaySchedule = getScheduleForWeekday(nextDay);

        if (state == ScheduleState.open ||
            state == ScheduleState.openAllDay && weekdaySchedule.firstOpen != null) {
          return DateHelper.timeStringToDateTime(
            weekdaySchedule.firstOpen!,
            nextDay,
          );
        }
      }

      return todaySchedule.opensAt(today);
    }

    return todaySchedule.opensAt(today);
  }

  DateTime? getNextClosingTime([DateTime? clock]) {
    final today = clock ?? DateTime.now();
    final todaySchedule = getScheduleForWeekday(today);
    final isCurrentlyOpen = todaySchedule.isOpen(today);

    if (isCurrentlyOpen) {
      return todaySchedule.closesAt(today);
    }

    for (var day in days) {
      if (day == todaySchedule) continue;

      final nextDay = today.add(
        const Duration(days: 1),
      );

      final weekdayDate = getScheduleForWeekday(nextDay);

      if (day.opensThisDate(nextDay) && weekdayDate.firstClose != null) {
        return DateHelper.timeStringToDateTime(
          weekdayDate.firstClose!,
          nextDay,
        );
      }
    }

    return null;
  }

  Schedule clone() {
    return Schedule.fromJson(toJson());
  }

  /// Runs a map function against each [ScheduleDay] in [days]
  void updateEachDay(ScheduleDay Function(ScheduleDay) toElement) {
    days = days.map(toElement).toList();
  }

  factory Schedule.fromJsonString(String jsonString) {
    var parsedJson = _tryParseScheduleJsonString(jsonString)!;
    return Schedule.fromJson(parsedJson);
  }

  factory Schedule.fromJson(Map<String, dynamic> json) {
    var days = <ScheduleDay>[];

    if (json['days'] != null) {
      var jsonWallets = List.from(json['days']);
      days = jsonWallets.isNotEmpty
          ? jsonWallets.map<ScheduleDay>((el) => ScheduleDay.fromJson(el)).toList()
          : <ScheduleDay>[];
    }

    return Schedule(
      type: ScheduleType.fromName(json['type'] ?? 'not_available'),
      days: days,
    );
  }

  Map<String, dynamic> toJson() {
    return {
      'type': type!.type,
      'days': days.map((e) => e.toJson()).toList(),
    };
  }

  static Map<String, dynamic>? _tryParseScheduleJsonString(String jsonString) {
    try {
      var parsedJson = json.decode(jsonString);
      return parsedJson;
    } catch (e) {
      return {};
    }
  }
}
